'use strict'

const {
	stringify,
	appendToUrl: appendQuerystring,
} = require('http-querystring-stringify')
const parseJson = require('json-parse-safe')

class TokenExpiredError extends Error {
	constructor(head, body) {
		super('Access token expired')
		this.name = this.constructor.name
		this.response = head
		this.body = body
	}
}

class RateLimitError extends Error {
	constructor(head, body) {
		super('Rate limited')
		this.name = this.constructor.name
		this.response = head
		this.body = body
	}
}

class UnexpectedHttpStatusError extends Error {
	constructor(head, body) {
		super(`HTTP error: ${head.status} (${head.statusText})`)
		this.name = this.constructor.name
		this.status = head.status
		this.response = head
		this.body = body
	}
}

class UnexpectedHttpContentTypeError extends Error {
	constructor(head, body) {
		super(`Unexpected Content-Type: ${head.headers.get('content-type')}`)
		this.name = this.constructor.name
		this.response = head
		this.body = body
	}
}

class InvalidJsonBodyError extends Error {
	constructor(head, body) {
		super(`Unexpected Content-Type: ${head.headers.get('content-type')}`)
		this.name = this.constructor.name
		this.response = head
		this.body = body
	}
}

pfetch.TokenExpiredError = TokenExpiredError
pfetch.RateLimitError = RateLimitError
pfetch.UnexpectedHttpStatusError = UnexpectedHttpStatusError
pfetch.UnexpectedHttpContentTypeError = UnexpectedHttpContentTypeError
pfetch.InvalidJsonBodyError = InvalidJsonBodyError
pfetch.formatPeriod = formatPeriod

module.exports = pfetch

function pfetch({
	fetch = globalThis.fetch,
	log = () => {},
	host = 'https://api.podio.com',
	token,
	method,
	path = '/',
	query = {},
	form,
	json,
}) {
	const hasForm = form !== undefined
	const hasJson = json !== undefined
	if (hasForm && hasJson) {
		throw new Error('`form` and `json` parameters are mutually exclusive')
	}
	if (method === undefined) {
		method = hasForm || hasJson ? 'POST' : 'GET'
	}
	const url = appendQuerystring(`${host}${path}`, query)
	log({
		t: 'fetch',
		token,
		method,
		url,
		form,
		json,
	})
	const contentType = inferContentType({form, json})
	const head = fetch(url, {
		method,
		headers: {
			...(token && {authorization: `OAuth2 ${token}`}),
			...(contentType && {'content-type': contentType}),
		},
		body: compileBody({form, json}),
	}).then((head) => {
		log({
			t: 'fetched',
			status: head.status,
			podioRequestId: head.headers.get('x-podio-request-id'),
		})
		return head
	})
	const body = head.then((head) => head.text())
	return Promise.all([head, body]).then(([head, bodyText]) => {
		const {ok, status, headers} = head
		const {error: bodyError, body} = parseBody(head, bodyText)
		if (status === 420) {
			return Promise.reject(new RateLimitError(head, body))
		} else if (
			status === 401 &&
			headers.get('authorization') !== null &&
			headers.get('authorization').includes('expired_token')
		) {
			return Promise.reject(new TokenExpiredError(head, body))
		} else if (!ok) {
			return Promise.reject(new UnexpectedHttpStatusError(head, body))
		} else if (!hasJsonContentTypeHeader(head, body)) {
			return Promise.reject(
				new UnexpectedHttpContentTypeError(head, body)
			)
		} else if (bodyError !== undefined) {
			return Promise.reject(new InvalidJsonBodyError(head, bodyText))
		} else {
			return body
		}
	})
}

function parseBody(head, body) {
	if (hasJsonContentTypeHeader(head, body)) {
		const {error, value} = parseJson(body)
		return {error, body: value}
	} else {
		return {error: undefined, body}
	}
}

function hasJsonContentTypeHeader(head) {
	return (
		typeof head.headers.get('content-type') === 'string' &&
		head.headers.get('content-type').includes('json')
	)
}

function formatPeriod(start, end, utc = false) {
	if (!(start instanceof Date))
		throw new Error('`start` must be a Date object')
	if (end !== undefined && end !== null && !(end instanceof Date))
		throw new Error('`end` must be a Date object')

	return {
		start: formatDate(start, utc),
		end: end ? formatDate(end, utc) : undefined,
	}
}

function formatDate(d, utc) {
	if (utc) return d.toISOString().substring(0, 19).replace('T', ' ')

	const pad = (v) => (v + '').padStart(2, '0')

	const year = d.getFullYear()
	const month = pad(d.getMonth() + 1)
	const day = pad(d.getDate())
	const hours = pad(d.getHours())
	const minutes = pad(d.getMinutes())
	const seconds = pad(d.getSeconds())

	return `${year}-${month}-${day} ${hours}:${minutes}:${seconds}`
}

function inferContentType({form, json}) {
	if (form !== undefined) return 'application/x-www-form-urlencoded'
	if (json !== undefined) return 'application/json'
	return undefined
}

function compileBody({form, json}) {
	if (form !== undefined) return stringify(form)
	if (json !== undefined) return JSON.stringify(json)
	return undefined
}
