'use strict'

const http = require('node:http')
const {once} = require('events')
const {join} = require('bluebird')
const test = require('tape')
const fetch = require('./')

const server = http
	.createServer((req, res) => {
		switch (req.url.split('?')[0]) {
			case '/':
			case '/echo': {
				let body = ''
				res.writeHead(200, {
					'Content-Type': 'application/json; charset=utf-8',
					'x-podio-request-id': 'PODIO_REQUEST_ID',
				})
				req.setEncoding('utf8')
				req.on('data', (chunk) => (body += chunk))
				req.on('end', () => {
					res.end(
						JSON.stringify({
							method: req.method,
							path: req.url,
							headers: req.headers,
							body,
						})
					)
				})
				break
			}
			case '/rate-limit': {
				res.writeHead(420)
				res.end()
				break
			}
			case '/expired-token': {
				res.writeHead(401, {
					authorization:
						'Bearer realm="podio", error="expired_token"',
				})
				res.end()
				break
			}
			case '/500': {
				res.writeHead(500)
				res.end()
				break
			}
			case '/400-json': {
				res.writeHead(400, {
					'Content-Type': 'application/json; charset=utf-8',
				})
				res.end(JSON.stringify({key: 'value'}))
				break
			}
			case '/400-text': {
				res.writeHead(400, {
					'Content-Type': 'text/plain; charset=utf-8',
				})
				res.end('text')
				break
			}
			case '/text': {
				res.writeHead(200, {'content-type': 'text/plain'})
				res.end()
				break
			}
			case '/invalid-json': {
				res.writeHead(200, {'content-type': 'application/json'})
				res.end('Invalid JSON')
				break
			}
			default:
				throw new Error('Oops')
		}
	})
	.listen(0)

const address = once(server, 'listening').then(
	() => `http://127.0.0.1:${server.address().port}`
)

test('minimal request', (t) => {
	t.plan(4)
	const response = join(address, (address) => fetch({host: address}))
	join(response, ({method, path, headers, body}) => {
		t.equal(method, 'GET')
		t.equal(path, '/')
		t.deepEqual(headers, defaultHeaders(headers))
		t.equal(body, '')
	})
})

test('with a token', (t) => {
	t.plan(4)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/echo',
			token: 'some-token',
		})
	)
	join(response, ({method, path, headers, body}) => {
		t.equal(method, 'GET')
		t.equal(path, '/echo')
		t.deepEqual(headers, {
			...defaultHeaders(headers),
			authorization: 'OAuth2 some-token',
		})
		t.equal(body, '')
	})
})

test('with a query', (t) => {
	t.plan(4)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/echo',
			query: {key: 'value', ghost: undefined},
		})
	)
	join(response, ({method, path, headers, body}) => {
		t.equal(method, 'GET')
		t.equal(path, '/echo?key=value')
		t.deepEqual(headers, defaultHeaders(headers))
		t.equal(body, '')
	})
})

test('with a form', (t) => {
	t.plan(4)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/echo',
			form: {key: 'value'},
		})
	)
	join(response, ({method, path, headers, body}) => {
		t.equal(method, 'POST')
		t.equal(path, '/echo')
		t.deepEqual(headers, {
			...defaultHeaders(headers),
			'content-type': 'application/x-www-form-urlencoded',
			'content-length': '9',
		})
		t.equal(body, 'key=value')
	})
})

test('with JSON', (t) => {
	t.plan(4)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/echo',
			json: {key: 'value'},
		})
	)
	join(response, ({method, path, headers, body}) => {
		t.equal(method, 'POST')
		t.equal(path, '/echo')
		t.deepEqual(headers, {
			...defaultHeaders(headers),
			'content-type': 'application/json',
			'content-length': '15',
		})
		t.equal(body, '{"key":"value"}')
	})
})

test('both JSON and form body', (t) => {
	t.plan(1)
	t.throws(
		() =>
			fetch({
				json: true,
				form: {key: true},
			}),
		/`form` and `json` parameters are mutually exclusive/
	)
})

test('rate limited', (t) => {
	t.plan(1)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/rate-limit',
		})
	)
	response.then(
		() => t.fail(),
		(err) => t.ok(err instanceof fetch.RateLimitError)
	)
})

test('expired token', (t) => {
	t.plan(1)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/expired-token',
		})
	)
	response.then(
		() => t.fail(),
		(err) => t.ok(err instanceof fetch.TokenExpiredError)
	)
})

test('unexpected status code', (t) => {
	t.plan(2)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/500',
		})
	)
	response.then(
		() => t.fail(),
		(err) => {
			t.ok(err instanceof fetch.UnexpectedHttpStatusError)
			t.equal(err.status, 500)
		}
	)
})

test('unexpected content type', (t) => {
	t.plan(1)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/text',
		})
	)
	response.then(
		() => t.fail(),
		(err) => t.ok(err instanceof fetch.UnexpectedHttpContentTypeError)
	)
})

test('invalid JSON body', (t) => {
	t.plan(2)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/invalid-json',
		})
	)
	response.then(
		() => t.fail(),
		(err) => {
			t.ok(err instanceof fetch.InvalidJsonBodyError)
			t.deepEqual(err.body, 'Invalid JSON')
		}
	)
})

test('error with JSON body', (t) => {
	t.plan(3)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/400-json',
		})
	)
	response.then(
		() => t.fail(),
		(err) => {
			t.ok(err instanceof fetch.UnexpectedHttpStatusError)
			t.equal(err.status, 400)
			t.deepEqual(err.body, {key: 'value'})
		}
	)
})

test('error with text body', (t) => {
	t.plan(3)
	const response = join(address, (address) =>
		fetch({
			host: address,
			path: '/400-text',
		})
	)
	response.then(
		() => t.fail(),
		(err) => {
			t.ok(err instanceof fetch.UnexpectedHttpStatusError)
			t.equal(err.status, 400)
			t.deepEqual(err.body, 'text')
		}
	)
})

test('invalid host', (t) => {
	t.plan(2)
	fetch({
		host: 'http://0.0.0.0',
		path: '/',
	}).then(
		() => t.fail(),
		(err) => {
			t.ok(err instanceof TypeError)
			t.equal(err.cause.code, 'ECONNREFUSED')
		}
	)
})

test('`formatPeriod`', (t) => {
	t.plan(6)

	const start = new Date('2020-01-21T10:30:50.000Z')
	const end = new Date('2020-01-22T10:30:50.000Z')

	t.throws(
		() => fetch.formatPeriod('2020-01-21T10:30:50.000Z'),
		/`start` must be a Date object/,
		'Validates type of `start`'
	)
	t.throws(
		() => fetch.formatPeriod(start, '2020-01-21T10:30:50.000Z'),
		/`end` must be a Date object/,
		'Validates type of `end`'
	)

	const pad = (v) => (v + '').padStart(2, '0')

	t.deepEqual(
		fetch.formatPeriod(start),
		{
			start: `2020-01-${pad(start.getDate())} ${pad(
				start.getHours()
			)}:30:50`,
			end: undefined,
		},
		'Formats the period in Date timezone'
	)
	t.deepEqual(
		fetch.formatPeriod(start, end),
		{
			start: `2020-01-${pad(start.getDate())} ${pad(
				start.getHours()
			)}:30:50`,
			end: `2020-01-${pad(end.getDate())} ${pad(end.getHours())}:30:50`,
		},
		'Formats the period in Date timezone'
	)

	t.deepEqual(
		fetch.formatPeriod(start, null, true),
		{
			start: `2020-01-${pad(start.getUTCDate())} ${pad(
				start.getUTCHours()
			)}:30:50`,
			end: undefined,
		},
		'Formats the period in UTC'
	)
	t.deepEqual(
		fetch.formatPeriod(start, end, true),
		{
			start: `2020-01-${pad(start.getUTCDate())} ${pad(
				start.getUTCHours()
			)}:30:50`,
			end: `2020-01-${pad(end.getUTCDate())} ${pad(
				end.getUTCHours()
			)}:30:50`,
		},
		'Formats the period in UTC'
	)
})

test('teardown', (t) => {
	server.close()
	t.end()
})

function defaultHeaders(headers) {
	return {
		host: headers.host,
		connection: 'keep-alive',
		accept: '*/*',
		'accept-language': '*',
		'sec-fetch-mode': 'cors',
		'user-agent': 'undici',
		'accept-encoding': 'gzip, deflate',
	}
}
